from collections import OrderedDict as _ODict


def check(logic, k, v, kw):
    phases = []
    for site_vars in kw.values():
        phases.append(site_vars.get(k) == v)
    return logic(phases)


class FrozenDict(dict):
    def __init__(self, _dict):
        super().__init__(_dict)

    def prompt(self, key, value):
        raise ValueError(f'*** '
                         f'Attempt to modify frozen dict '
                         f'[{key} : {self[key]}] with [{key} : {value}]'
                         f' ***')

    def __setitem__(self, key, value):
        if key not in self:
            super(FrozenDict, self).__setitem__(key, value)
        else:
            self.prompt(key, value)

    def update(self, **kw):
        for k, v in kw.items():
            self[k] = v


class ComputationPhase:
    def __init__(self, phase_id, cache, input, state, **kw):
        self.id = f"PHASE:{phase_id}"
        self.cache = cache
        self.input = input
        self.state = state

        "Cached default input obtained from inputspec.json file to reuse during multiple iterations"
        self.input_args = cache['input_args']

        if not self.cache.get(self.id):
            self._initialize()
            self.cache[self.id] = True

    def _initialize(self):
        """Runs only once"""
        pass

    def compute(self):
        return {}

    def __str__(self):
        return f"{self.id}"


class PhaseEndWithSuccess(ComputationPhase):
    def compute(self):
        return {'success': True}


class PhasePipeline:
    def __init__(self, pipeline_id, cache, **kw):
        self.id = f"PIPELINE:{pipeline_id}"
        self.cache = cache

        if not cache.get(self.id):
            self._initialize()

        self.phases = _ODict()
        self.multi_iterations = {}

    def _initialize(self):
        """Runs only once"""
        self.cache[self.id] = {'index': 0, 'iterations': {}}

    def add_phase(self, phase_cls, multi_iterations=False):
        self.phases[phase_cls.__name__] = phase_cls
        self.multi_iterations[phase_cls.__name__] = multi_iterations
        self.cache[self.id]['iterations'][phase_cls.__name__] = 0

    def __len__(self):
        return len(self.phases)

    @property
    def phase_ids(self):
        return list(self.phases.keys())

    def next_phase(self, jump_phase=False):
        current_phase_key = self.phase_ids[self.cache[self.id]['index']]

        _is_single_iteration_phase = not self.multi_iterations[current_phase_key]
        if jump_phase or _is_single_iteration_phase:
            if self.cache[self.id]['index'] < len(self.phases) - 1:
                self.cache[self.id]['index'] += 1

        self.cache[self.id]['iterations'][self.phase_ids[self.cache[self.id]['index']]] += 1
        return self.phase_ids[self.cache[self.id]['index']]

    def __str__(self):
        return [f"{id}" for id in self.phases]



import copy as _copy
import json as _json
import sys as _sys
import traceback as _tb

import coinstac_computation.utils as _utils


class COINSTACPyNode:
    _VALID_MODES_ = ['REMOTE', 'LOCAL']

    def __init__(self, mode: str, debug=False):
        self._cache = {}
        self._mode = mode.upper()
        self._logs = {}
        self._debug = debug
        self._pipeline = _utils.PhasePipeline(self._mode, self._cache)

        assert self._mode in COINSTACPyNode._VALID_MODES_, \
            f"Invalid mode : {self._mode}, Use one of: {COINSTACPyNode._VALID_MODES_}"

    def add_phase(self, phase_cls, multi_iterations=False):
        self._pipeline.add_phase(phase_cls, multi_iterations)

    def _print_logs(self):
        print()
        for k, v in self._logs.items():
            print(f"[{k.upper()}]{'-' * 51}")
            for k1, v1 in v.items():
                print(f"\t[ {k1} ] -> {v1}")

    def compute(self, data):
        out = {}
        if self._debug:
            self._logs['input'] = {}
            self._logs['input']['PRE-COMPUTATION '] = _copy.deepcopy(data)
            self._logs['cache'] = {}
            self._logs['cache']['PRE-COMPUTATION '] = _copy.deepcopy(self._cache)

        if not self._cache.get('input_args'):
            """Some multi-iteration computations might need to reuse initial parameters, so save it."""
            self._cache['input_args'] = _utils.FrozenDict(_copy.deepcopy(data['input']))

        phase_key = self._cache.get('next_phase')
        if not phase_key:
            phase_key = self._pipeline.next_phase()

        phase = self._pipeline.phases[phase_key](
            phase_key,
            self._cache, data['input'], data['state']
        )

        try:
            _out = phase.compute()
            if _out:
                out.update(**_out)
        except:
            raise RuntimeError(f"ERROR! in Phase: *** {phase_key} ***")

        output = {"output": out}

        if self._mode == 'REMOTE':
            output['success'] = out.get('success', False)

        self._cache['next_phase'] = self._pipeline.next_phase(
            out.get('jump_to_next', False)
        )

        if self._debug:
            self._logs['cache']['POST-COMPUTATION'] = _copy.deepcopy(self._cache)
            self._logs['output'] = {'POST-COMPUTATION': _copy.deepcopy(output)}
            self._print_logs()

        return output

    def to_stdout(self):
        """
        Backward compatibility for the old library. Deprecated now.
        """
        data = _json.loads(_sys.stdin.read())
        output = {}
        if data.get('cache') and len(data['cache']) > 0:
            self._cache = data['cache']
            self._pipeline.cache = self._cache
        try:
            self._debug = False
            output = self.compute(data)
            output['cache'] = self._cache
            output = _json.dumps(output)
            _sys.stdout.write(output)
        except Exception as e:
            _tb.print_exc()
            raise Exception(f"{data['state']['clientId']} error {e} \n ### Output: {output} ***")
